from ast import *
from aporiapy.compilers.utils import generate_name


class CompilerSc:

    def compile(self, program_ast: Module) -> Module:

        return Module(self.rco_statements(program_ast.body), [])

    def rco_statements(self, stmts):

        new_stmts = []

        for statement in stmts:
            new_stmts.extend(self.rco_statement(statement))
        return new_stmts

    def rco_statement(self, stmt) -> list[stmt]:
        new_stmts = []
        match stmt:
            case If(test, body, orelse):
                new_test, temp = self.rco_expression(test, True)
                new_stmts.extend(temp)
                neg_test = Name(generate_name())
                new_stmts.append(Assign([neg_test], UnaryOp(Not(), new_test)))
                new_body = self.rco_statements(body)
                new_orelse = self.rco_statements(orelse)
                new_stmts.append(If(new_test, new_body, []))
                new_stmts.append(If(neg_test, new_orelse, []))
                return new_stmts
            case Expr(Call(Name("print"), [exp])):
                exp, temps = self.rco_expression(exp, False)
                new_stmts.extend(temps)
                new_stmts.append(Expr(Call(Name("print"), [exp], keywords=[])))
                return new_stmts
            case Expr():
                return []
            case Assign([Name(var)], expr):
                return self.rco_assign(var, expr)
            case _:
                raise Exception(f"Unexpected statement in rco_statement: {stmt}")

    def rco_assign(self, name, expr) -> list[stmt]:
        new_stmts = []
        match expr:
            case IfExp(test, body, orelse):
                return self.rco_if_exp(name, test, body, orelse)

            case _:
                expr, temps = self.rco_expression(expr, False)
                new_stmts.extend(temps)
                new_stmts.append(Assign([Name(name)], expr))
                return new_stmts

    def rco_if_exp(self, name, test, body, orelse) -> list[stmt]:
        new_stmts = []
        new_test, temp = self.rco_expression(test, True)
        new_stmts.extend(temp)
        neg_test = Name(generate_name())
        new_stmts.append(Assign([neg_test], UnaryOp(Not(), new_test)))
        body_expr, body_temps = self.rco_expression(body, False)
        orelse_expr, orelse_temps = self.rco_expression(orelse, False)
        body_assign = [Assign([Name(name)], body_expr)]
        orelse_assign = [Assign([Name(name)], orelse_expr)]
        new_body = body_temps + body_assign
        new_orelse = orelse_temps + orelse_assign
        new_stmts.append(If(new_test, new_body, []))
        new_stmts.append(If(neg_test, new_orelse, []))
        return new_stmts

    def rco_expression(self, expression, need_atomic) -> tuple[expr, list[stmt]]:

        def make_atom_if_needed(exp, exp_temps) -> tuple[expr, list[stmt]]:
            if need_atomic:
                temp_var = Name(generate_name())
                exp_temps.append(Assign([temp_var], exp))
                exp = temp_var
            return exp, exp_temps

        match expression:
            case IfExp(test, body, orelse):
                name = generate_name("temp")
                temps = self.rco_if_exp(name, test, body, orelse)
                return Name(name), temps
            case BinOp(left, op, right):
                left, left_temps = self.rco_expression(left, False)
                right, right_temps = self.rco_expression(right, False)
                exp = BinOp(left, op, right)
                exp_temps = left_temps + right_temps
                return make_atom_if_needed(exp, exp_temps)
            case UnaryOp(op, exp):
                exp, exp_temps = self.rco_expression(exp, False)
                exp = UnaryOp(op, exp)
                return make_atom_if_needed(exp, exp_temps)
            case Constant() | Name():
                return expression, []
            case BoolOp(op, [left, right]):
                left, left_temps = self.rco_expression(left, False)
                right, right_temps = self.rco_expression(right, False)
                exp = BoolOp(op, [left, right])
                exp_temps = left_temps + right_temps
                return make_atom_if_needed(exp, exp_temps)
            case Compare(left, cmp, [right]):
                left, left_temps = self.rco_expression(left, False)
                right, right_temps = self.rco_expression(right, False)
                exp = Compare(left, cmp, [right])
                exp_temps = left_temps + right_temps
                return make_atom_if_needed(exp, exp_temps)
            case _:
                raise Exception(
                    f"Unexpected expression in rco_expression: {expression}"
                )



import argparse
import ast
import sys
from pathlib import Path

from aporiapy.compilers.compiler_01_sc import CompilerSc
from aporiapy.compilers.compiler_02_flat import CompilerFlat
from aporiapy.compilers.compiler_03_cfi import CompilerCfi

compilers = [CompilerSc(), CompilerFlat(), CompilerCfi()]


def compile_source(input_path, output_to_stdout, output_path):
    """Compile the source file and handle output based on the mode."""
    with open(input_path, "r") as file:
        original_source = file.read()

    program_ast = ast.parse(original_source)

    compiled_source = original_source
    for compiler in compilers:
        compiled_ast = compiler.compile(program_ast)
        if isinstance(compiled_ast, ast.Module):
            ast.fix_missing_locations(compiled_ast)
            compiled_source = ast.unparse(compiled_ast)
        else:
            compiled_source = str(compiled_ast)

        program_ast = compiled_ast

    if output_to_stdout:
        print(compiled_source)
    else:
        if output_path is None:
            output_path = input_path.with_suffix(".spp")
        with open(output_path, "w") as output_file:
            output_file.write(compiled_source)
        print(f"Compiled file written to: {output_path}")

def cli():
    parser = argparse.ArgumentParser(description="Compiles Python code to Aporia code")

    parser.add_argument(
        "input_file", type=str, help="Path to the input Python file to compile."
    )

    parser.add_argument(
        "-o", "--output", type=str, help="Path to the output file to write the compiled code."
    )

    parser.add_argument(
        "--stdout", action="store_true", help="Output the compiled code to stdout instead of a file."
    )

    args = parser.parse_args()

    input_path = Path(args.input_file)

    if not input_path.exists():
        sys.stderr.write(f"Error: Input file '{input_path}' does not exist.\n")
        sys.exit(1)

    compile_source(input_path, args.stdout, args.output)

if __name__ == "__main__":
    cli()





import ast
from ast import *
import aporia.aporia_ast as lcfi_ast
from aporia.aporia_ast import *


class CompilerCfi:

    def compile(self, program_ast: Module) -> L_cfi:
        int_vars = set()
        bool_vars = set()
        ap_stmts = []
        for stmt in program_ast.body:
            if not isinstance(stmt, ast.If):
                raise Exception("Only if statements are allowed in the program")
            ap_stmts.append(self.translate_stmt(stmt, int_vars, bool_vars))

        int_vars = {Var(var) for var in int_vars}
        bool_vars = {Var(var) for var in bool_vars}
        ap_declar = []
        if not len(int_vars) == 0:
            ap_declar.append(Declar(Int(), int_vars))
        if not len(bool_vars) == 0:
            ap_declar.append(Declar(Bool(), bool_vars))
        cfi = L_cfi(ap_declar, ap_stmts)
        return cfi

    def translate_stmt(self, if_stmt, int_vars, bool_vars) -> Stmt:
        test = if_stmt.test
        ap_test = self.select_pred(test)
        body = if_stmt.body
        if not body or len(body) != 1:
            raise Exception("Body of if statement must have exactly one statement")
        ap_body = self.select_instruction(body[0], int_vars, bool_vars)
        return Stmt(None, ap_test, ap_body)

    def select_pred(self, test) -> Pred:
        match test:
            case ast.Name(var):
                return Pred(Var(var))
            case ast.Constant(value):
                return Pred(Bools(value))
            case _:
                raise Exception(f"Unexpected test in select_pred: {test}")

    def select_instruction(self, stmt, int_vars, bool_vars) -> Inst:
        match stmt:
            case ast.Expr(Call(Name('print'), [exp])):
                return PrintInst("", self.select_exp(exp))
            case ast.Assign([Name(var)], exp):
                self.check_type(var, exp, int_vars, bool_vars)
                return AssignInst(lcfi_ast.Assign(Var(var), self.select_exp(exp)))
            case ast.Expr(value):
                return ExpInst(self.select_exp(value))
            case _:
                raise Exception(f"Unexpected statement in select_instruction: {stmt}")


    def check_type(self, new_var, exp, int_vars, bool_vars):
        match exp:
            case ast.Name(var):
                if var in int_vars:
                    int_vars.add(new_var)
                elif var in bool_vars:
                    bool_vars.add(new_var)
                else:
                    raise Exception(f"Variable {var} not declared")
            case ast.BinOp():
                int_vars.add(new_var)
            case ast.BoolOp():
                bool_vars.add(new_var)
            case ast.UnaryOp(ast.USub(), _):
                int_vars.add(new_var)
            case ast.UnaryOp(ast.Not(), _):
                bool_vars.add(new_var)
            case ast.Compare():
                bool_vars.add(new_var)
            case ast.Constant(value):
                if isinstance(value, bool):
                    bool_vars.add(new_var)
                elif isinstance(value, int):
                    int_vars.add(new_var)
            case _:
                raise Exception(f"Unexpected expression in check_type: {exp}")

    def select_exp(self, exp) -> Exp:
        match exp:
            case ast.Name(var):
                return Var(var)
            case ast.Constant(value):
                return lcfi_ast.Bools(value) if isinstance(value,bool) else lcfi_ast.Constant(value)
            case ast.BinOp(left, op, right):
                op = self.select_op(op)
                return lcfi_ast.BinOp(self.select_exp(left), op, self.select_exp(right))
            case ast.UnaryOp(op, operand):
                op = self.select_op(op)
                return lcfi_ast.UnaryOp(op, self.select_exp(operand))
            case ast.BoolOp(op, [left, right]):
                op = self.select_op(op)
                return lcfi_ast.BinOp(self.select_exp(left), op, self.select_exp(right))
            case ast.Compare(left, [cmp], [right]):
                cmp = self.select_cmp(cmp)
                return lcfi_ast.BinOp(self.select_exp(left), cmp, self.select_exp(right))
            case _:
                raise Exception(f"Unexpected expression in select_exp: {exp}")

    def select_op(self, op) -> Operator:
        match op:
            case ast.Add():
                return lcfi_ast.Add()
            case ast.Sub():
                return lcfi_ast.Sub()
            case ast.USub():
                return lcfi_ast.USub()
            case ast.And():
                return lcfi_ast.And()
            case ast.Or():
                return lcfi_ast.Or()
            case ast.Not():
                return lcfi_ast.Not()
            case _:
                raise Exception(f"Unexpected operator in select_op: {op}")

    def select_cmp(self, cmp) -> Comparator:
        match cmp:
            case ast.Eq():
                return lcfi_ast.Eq()
            case ast.NotEq():
                return lcfi_ast.Neq()
            case ast.Lt():
                return lcfi_ast.Lt()
            case ast.LtE():
                return lcfi_ast.Le()
            case ast.Gt():
                return lcfi_ast.Gt()
            case ast.GtE():
                return lcfi_ast.Ge()
            case _:
                raise Exception(f"Unexpected comparator in select_cmp: {cmp}")




from ast import (
    Module,
    stmt,
    If,
    BoolOp,
    And,
    Constant,
    Name,
    Assign,
)

from aporiapy.compilers.utils import generate_name


class CompilerFlat:

    def compile(self, program_ast: Module) -> Module:
        out_list = []
        self.flatten_statements(program_ast.body, out_list)
        return Module(out_list, [])

    def flatten_statements(self, stmts: list[stmt], out_list: list[stmt], cond=None):

        for statement in stmts:
            self.flatten_statement(statement, out_list, cond)

    def flatten_statement(self, statement: stmt, out_list: list[stmt], cond=None):
        match statement:
            case If(test, body, []):
                if not cond:
                    temp = test
                else:
                    temp = Name(generate_name())
                    new_cond = BoolOp(And(), [cond, test])
                    out_list.append(If(Constant(True), [Assign([temp], new_cond)], []))
                for stmt in body:
                    match stmt:
                        case If():
                            self.flatten_statement(stmt, out_list, temp)
                        case _:
                            out_list.append(If(temp, [stmt], []))
            case _:
                out_list.append(If(Constant(True), [statement], []))
